(******************************************************************************)
(** * General memory subsystem definitions and simulation arguments   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Labels.

Set Implicit Arguments.

Section MS.

Variable State : Type.
Variable MS_init : State -> Prop.
Variable MS_step : State -> State -> option (prod Tid Lab)-> Prop.

(******************************************************************************)
(** * Three equivalent definitions of a run   *)
(******************************************************************************)

Inductive MS_run_from_to_state_1n : 
  list (prod Tid Lab) -> State ->  State -> Prop := 
  | empty_1n : forall st, MS_run_from_to_state_1n nil st st
  | step_1n  : forall t l L L' st0 st st', L = (t,l) :: L' -> 
                MS_run_from_to_state_1n L' st st' -> 
                MS_step st0 st (Some (t,l)) -> 
                MS_run_from_to_state_1n L st0 st'
  | eps_1n   : forall L st0 st st', 
                MS_run_from_to_state_1n L st st' -> 
                MS_step st0 st None -> 
                MS_run_from_to_state_1n L st0 st'.

Inductive MS_run_from_to_state_n1 : 
  list (prod Tid Lab) -> State ->  State -> Prop := 
  | empty_n1 : forall st, MS_run_from_to_state_n1 nil st st
  | step_n1  : forall t l L L' st0 st st', 
                L = L' ++ (cons (t,l) nil)-> 
                MS_run_from_to_state_n1 L' st0 st -> 
                MS_step st st' (Some (t,l)) -> 
                MS_run_from_to_state_n1 L st0 st'
  | eps_n1   : forall L st0 st st', 
                MS_run_from_to_state_n1 L st0 st -> 
                MS_step st st' None -> 
                MS_run_from_to_state_n1 L st0 st'.

Inductive MS_run_from_to_state : 
  list (prod Tid Lab) -> State ->  State -> Prop := 
  | empty    : forall st, MS_run_from_to_state nil st st
  | eps      : forall st st', 
                MS_step st st' None -> 
                MS_run_from_to_state nil st st'
  | step      : forall st st' t l, 
                MS_step st st' (Some (t,l)) -> 
                MS_run_from_to_state ((t,l)::nil) st st'
  | trans      : forall st st' st'' L L', 
                MS_run_from_to_state L st st' -> 
                MS_run_from_to_state L' st' st'' -> 
                MS_run_from_to_state (L ++ L') st st''.


(******************************************************************************)
(** * Equivalence of the three run definitions   *)
(******************************************************************************)

Lemma t1n_trans L st1 st2: 
 MS_run_from_to_state_1n L st1 st2 -> MS_run_from_to_state L st1 st2.
Proof.
ins; induction H.
- econs; eauto.
- subst L.
replace ((t, l) :: L') with ((cons (t, l) nil) ++ L').
2: by ins.
eapply trans; try edone.
econs; eauto.
- replace (L) with (nil ++ L).
2: by ins.
eapply trans; try edone.
econs; eauto.
Qed.

Lemma trans_t1n L st1 st2: 
 MS_run_from_to_state L st1 st2 -> MS_run_from_to_state_1n L st1 st2.
Proof.
ins; induction H.
- econs; eauto.
- eapply eps_1n; [econs|]; eauto.
- eapply step_1n; ins; [econs|]; eauto.
- revert IHMS_run_from_to_state2.
induction IHMS_run_from_to_state1; ins; try subst L.
+ rewrite <- app_comm_cons.
econs; ins; try edone.
eapply IHIHMS_run_from_to_state1; try done.
by apply t1n_trans.
+ eapply eps_1n; ins; try edone.
eapply IHIHMS_run_from_to_state1; try done.
by apply t1n_trans.
Qed.

Lemma t1n_iff_trans L st1 st2: 
 MS_run_from_to_state_1n L st1 st2 <-> MS_run_from_to_state L st1 st2.
Proof.
split.
apply t1n_trans.
apply trans_t1n.
Qed.

Lemma tn1_trans L st1 st2: 
 MS_run_from_to_state_n1 L st1 st2 -> MS_run_from_to_state L st1 st2.
Proof.
ins; induction H.
- econs; eauto.
- subst L.
replace ((t, l) :: L') with ((cons (t, l) nil) ++ L').
2: by ins.
eapply trans; try edone.
econs; eauto.
- replace (L) with (L ++ nil).
eapply trans; try edone.
econs; eauto.
apply app_nil_r.
Qed.

Lemma trans_tn1 L st1 st2: 
 MS_run_from_to_state L st1 st2 -> MS_run_from_to_state_n1 L st1 st2.
Proof.
ins; induction H.
- econs; eauto.
- eapply eps_n1; [econs|]; eauto.
- eapply step_n1 with (L':=nil); ins; [econs|]; eauto.
- revert IHMS_run_from_to_state1.
induction IHMS_run_from_to_state2; ins; try subst L0.
+ by rewrite app_nil_r.
+ rewrite app_assoc.
econs; ins; try edone.
eapply IHIHMS_run_from_to_state2; try done.
by apply tn1_trans.
+ eapply eps_n1; ins; try edone.
eapply IHIHMS_run_from_to_state2; try done.
by apply tn1_trans.
Qed.

Lemma tn1_iff_trans L st1 st2: 
 MS_run_from_to_state_n1 L st1 st2 <-> MS_run_from_to_state L st1 st2.
Proof.
split.
apply tn1_trans.
apply trans_tn1.
Qed.

Lemma tn1_iff_t1n L st1 st2: 
 MS_run_from_to_state_n1 L st1 st2 <-> MS_run_from_to_state_1n L st1 st2.
Proof.
by rewrite tn1_iff_trans, t1n_iff_trans.
Qed.

(******************************************************************************)
(** * Reachable states   *)
(******************************************************************************)

Definition MS_run L := 
  exists st0 st, MS_init st0 /\ MS_run_from_to_state L st0 st.

Definition reachable st := 
  exists st0 L, MS_init st0 /\ MS_run_from_to_state L st0 st.

Lemma MS_reachable_preserves 
 st (REACH: reachable st) (P : State -> Prop) 
 (INIT: forall st, MS_init st -> P st) 
 (PRES: forall st st' o, P st -> MS_step st st' o -> P st') : 
  P st.
Proof.
red in REACH; desc.
apply trans_tn1 in REACH0.
induction REACH0; eauto.
Qed.

Lemma MS_reachable_extends st st' o
 (REACH: reachable st) (STEP : MS_step st st' o):
 reachable st'.
Proof.
unfold reachable in *; desc.
destruct o; try destruct p.
- eexists _,_; splits; eauto.
eapply trans. 
edone.
eapply step; edone.
- eexists _,_; splits; eauto.
eapply trans. 
edone.
eapply eps; edone.
Qed.

End MS.

(******************************************************************************)
(** * Forward simulation  *)
(******************************************************************************)

Section FWD_SIM.

Variable State1 : Type.
Variable MS_init1 : State1 -> Prop.
Variable MS_step1 : State1 -> State1 -> option (prod Tid Lab)-> Prop.
Variable State2 : Type.
Variable MS_init2 : State2 -> Prop.
Variable MS_step2 : State2 -> State2 -> option (prod Tid Lab)-> Prop.

Variable sim_rel : State1 -> State2 -> Prop.

Variable sim_rel_fwd_init : 
 forall (st1 : State1),
  MS_init1 st1 -> exists st2, MS_init2 st2 /\ sim_rel st1 st2.

Variable sim_rel_fwd : 
 forall (st1 st1' : State1) (st2 : State2) o,
  MS_step1 st1 st1' o ->
  sim_rel st1 st2 ->
  (exists st2', MS_step2 st2 st2' o /\ sim_rel st1' st2') \/
  (o = None /\ sim_rel st1' st2).

Lemma MS1_runs_in_MS2_runs_fwd_helper : forall L st10 st20 st1, 
  sim_rel st10 st20 ->
  MS_run_from_to_state MS_step1 L st10 st1 -> 
  exists st2, 
    MS_run_from_to_state MS_step2 L st20 st2 /\ 
    sim_rel st1 st2.
Proof.
setoid_rewrite <- tn1_iff_trans.
ins.
induction H0; ins.
- eexists; splits; eauto; econs.
- specialize (IHMS_run_from_to_state_n1 H); desc.
eapply sim_rel_fwd in H2; try edone; desf.
eexists; splits; eauto.
econs; eauto.
- specialize (IHMS_run_from_to_state_n1 H); desc.
eapply sim_rel_fwd in H1; try edone; desf; eauto.
exists st2'; splits; eauto.
eapply eps_n1; eauto.
Qed.

Lemma MS1_runs_in_MS2_runs_fwd : forall L, 
  MS_run MS_init1 MS_step1 L -> MS_run MS_init2 MS_step2 L.
Proof.
ins.
unfold MS_run in *; desc.
apply sim_rel_fwd_init in H; desc.
eapply MS1_runs_in_MS2_runs_fwd_helper in H0; try edone; desc.
eauto.
Qed.

End FWD_SIM.


(******************************************************************************)
(** * Backward simulation  *)
(******************************************************************************)

Section BWD_SIM.

Variable State1 : Type.
Variable MS_init1 : State1 -> Prop.
Variable MS_step1 : State1 -> State1 -> option (prod Tid Lab)-> Prop.
Variable State2 : Type.
Variable MS_init2 : State2 -> Prop.
Variable MS_step2 : State2 -> State2 -> option (prod Tid Lab)-> Prop.

Variable sim_rel : State1 -> State2 -> Prop.

Variable sim_rel_tot : forall st1 
  (REACH: reachable MS_init1 MS_step1 st1), 
  exists st2, sim_rel st1 st2.

Variable sim_rel_bwd_init : 
 forall st1 st2 (REL: sim_rel st1 st2) (INIT: MS_init1 st1),
 MS_init2 st2.

Variable sim_rel_bwd : 
 forall (st1 st1' : State1) (st2' : State2) o
  (REACH: reachable MS_init1 MS_step1 st1),
  MS_step1 st1 st1' o ->
  sim_rel st1' st2' ->
  (exists st2, MS_step2 st2 st2' o /\ sim_rel st1 st2) \/
  (o = None /\ sim_rel st1 st2').

Lemma MS1_runs_in_MS2_runs_bwd_helper L st1 st1' st2'
  (REACH: reachable MS_init1 MS_step1 st1)
  (SIM: sim_rel st1' st2')
  (RUN: MS_run_from_to_state MS_step1 L st1 st1') :
  exists st2, 
    MS_run_from_to_state MS_step2 L st2 st2' /\ 
    sim_rel st1 st2.
Proof.
setoid_rewrite <- t1n_iff_trans.
setoid_rewrite <- t1n_iff_trans in RUN.
generalize dependent REACH.
generalize dependent SIM.
induction RUN; ins.
- eexists; splits; eauto; econs.
- forward (apply IHRUN).
done.
eapply MS_reachable_extends; try edone.
ins; desc.
eapply sim_rel_bwd in H2; try edone; desf. 
eexists; splits; eauto; econs; eauto.
- forward (apply IHRUN).
done.
eapply MS_reachable_extends; try edone.
ins; desc.
eapply sim_rel_bwd in H1; try edone; desf; eauto.
exists st1; splits; eauto.
eapply eps_1n; eauto.
Qed.

Lemma MS1_runs_in_MS2_runs_bwd : forall L, 
  MS_run MS_init1 MS_step1 L -> MS_run MS_init2 MS_step2 L.
Proof.
ins.
unfold MS_run in *; desc.
forward (apply sim_rel_tot).
by red; eauto.
ins; desc.
eapply MS1_runs_in_MS2_runs_bwd_helper in H0; try edone; desc.
exists st1, st2; splits; eauto.
red; eexists _,_; splits; eauto.
apply empty.
Qed.

End BWD_SIM.