(******************************************************************************)
(** * SRAL memory subsystem   *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.

Require Import Basic.
Require Import Labels.
Require Import Filters.
Require Import Subseq.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section SRAL.

(******************************************************************************)
(** * Read options  *)
(******************************************************************************)

Inductive Opt := 
  | R_opt (tw : Tid) (x : Labels.Loc) (v : Val) (rmw_flag : bool).

Definition tid_opt o :=
  match o with
  | R_opt tw _ _ _ => tw
  end.

Definition loc_opt o :=
  match o with
  | R_opt _ x _ _ => x
  end.

Definition val_opt o :=
  match o with
  | R_opt _ _ v _ => v
  end.

Definition rmw_flag_opt o :=
  match o with
  | R_opt _ _ _ u => u
  end.

Definition default_opt := R_opt 0 X0 0 false.

(******************************************************************************)
(** * States  *)
(******************************************************************************)

Definition State := Tid -> list Opt -> Prop.

Definition wf_state (st : State) := 
  forall t, exists ol, st t ol.

(******************************************************************************)
(** * Steps  *)
(******************************************************************************)

Definition SRAL_read_step_helper (st st' : State)  t x v u :=
 ⟪ P: forall p (NEQ: p <> t) ol' (IN: st' p ol'), st p ol' ⟫ /\
 ⟪ T: exists tw, forall ol' (IN: st' t ol'), 
    st t ((R_opt tw x v u) :: ol') ⟫.

Definition Ind_choice := Tid -> list Opt -> list bool.

Definition wf_ind_choice_ol (ol : list Opt) (P : list bool)
                         (t : Tid) (x : Labels.Loc) (v : Val)  :=
  ⟪ P_LEN:    length P = length ol ⟫ /\
  ⟪ P_TID:    forall n (NTH_IN: nth n P false = true),
               tid_opt (nth n ol default_opt) = t⟫ /\
  ⟪ P_LOC:    forall n (NTH_IN: nth n P false = true),
               loc_opt (nth n ol default_opt) = x⟫ /\
  ⟪ P_VAL:    forall n (NTH_IN: nth n P false = true),
               val_opt (nth n ol default_opt) = v⟫.

Definition wf_ind_choice (I : Ind_choice) (st' : State) 
                         (t : Tid) (x : Labels.Loc) (v : Val)  :=
  forall p ol (IN: st' p ol), wf_ind_choice_ol ol (I p ol) t x v.

Definition SRAL_write_step_helper (st st' : State)  t x v I :=
 ⟪ T_LOC: forall ol (IN: st t ol) o (INo: In o ol), loc_opt o <> x ⟫ /\
 ⟪ NU: forall p ol (IN: st p ol) o (INo: In o ol)
              (LOC: loc_opt o = x), rmw_flag_opt o <> true ⟫ /\
 ⟪ I_WF: wf_ind_choice I st' t x v ⟫ /\
 ⟪ P: forall p ol' (IN: st' p ol'), st p (filter_ind ol' (I p ol')) ⟫ /\
 ⟪ T: forall p ol' (IN: st' p ol'), st t (filter_ind_pre ol' (I p ol')) ⟫.

Inductive SRAL_step : State -> State -> option (prod Tid Lab)-> Prop :=
| step_read st st' t x v u
  (READ: SRAL_read_step_helper st st' t x v u)
  (WF': wf_state st') 
  : SRAL_step st st' (Some (t, (Aload x v)))
| step_write st st' t x v I
  (WRITE: SRAL_write_step_helper st st' t x v I)
  (WF': wf_state st') 
   : SRAL_step st st' (Some (t, (Astore x v)))
| step_rmw st st' st'' t x vr vw I
  (READ: SRAL_read_step_helper st st' t x vr true)
  (WRITE: SRAL_write_step_helper st' st'' t x vw I)
  (WF'': wf_state st'') 
  : SRAL_step st st'' (Some (t, (Armw x vr vw)))
| step_lose (st st' : State) 
  (SUB : forall p ol' (IN: st' p ol'), 
     exists ol, st p ol /\ subseq ol' ol)
  (WF': wf_state st') 
  : SRAL_step st st' None.

Lemma SRAL_step_preserves_wf st st' o (WF: wf_state st) 
  (STEP: SRAL_step st st' o): wf_state st'.
Proof.
inversion STEP; subst; clear STEP; eauto.
Qed.

(******************************************************************************)
(** * Initial state  *)
(******************************************************************************)

Definition SRALinit (st : State) : Prop :=
 wf_state st /\ 
 forall p ol (IN: st p ol) n (LEN: n < length ol),
  tid_opt (nth n ol default_opt) = 0 /\ 
  val_opt (nth n ol default_opt) = 0.

Lemma SRALinit_wf st (INIT: SRALinit st) : wf_state st.
Proof.
by red in INIT; desc.
Qed.

Lemma SRALinit_exists : exists st, SRALinit st.
Proof.
exists (fun p ol => ol = nil).
red; splits.
- red; ins; eexists; eauto.
- ins; desf; ins; omega.
Qed.

End SRAL.

