(******************************************************************************)
(** * Subsequences: definitions and lemmas        *)
(** * Two alternative presentations:              *)
(** *  1. a recursive definiton                   *)
(** *  2. a witness definirion: there exists a monontone mapping         *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.
Require Import List.
Require Import Basic.

Require Import Filters.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section Subseq.

Variable A : Type.

Inductive subseq : list A -> list A -> Prop :=
| sub_nil (l : list A) : subseq nil l
| sub_fst (a : A) (l1 l2 : list A) 
          (SUB: subseq l1 l2) : subseq  (a :: l1) (a :: l2)
| sub_nfst (a : A) (l1 l2 : list A) 
          (SUB: subseq l1 l2) : subseq  l1 (a :: l2).

Lemma subseq_length (l1 l2 : list A) (SUB: subseq l1 l2) :
  length l1 <= length l2.
Proof.
induction SUB; ins; omega.
Qed.

Lemma subseq_In (l1 l2 : list A) (SUB: subseq l1 l2) a:
  In a l1 -> In a l2.
Proof.
induction SUB.
desf.
simpl; ins; desf; eauto.
simpl; ins; desf; eauto.
Qed.

Definition subseq_witness {A: Type} (l1 l2 : list A) (f : nat -> nat) := 
  << MON: forall n1 n2, n1 < n2 -> f n1 < f n2 >> /\
  << NTH: forall n, n < length l1 -> 
             nth_error l1 n = nth_error l2 (f n) >> /\
  << LEN: forall n, n < length l1 -> f n < length l2 >>.

Lemma subseq_mapping (l1 l2 : list A) (SUB: subseq l1 l2) :
  exists (f : nat -> nat), subseq_witness l1 l2 f.
Proof.
induction SUB; ins; desc.
- exists (fun n => n); red; splits; eauto.
  ins; omega.
  ins; omega.
- exists (fun n => match n with
                   | 0 => 0
                   | S n' => S (f n')
                   end).
  red in IHSUB; desf.
  red; splits; ins; desf; try omega.
  + eauto using lt_S_n, lt_n_S.
  + eauto using lt_S_n.
  + eauto using lt_S_n, lt_n_S.
- exists (fun n => S (f n)).
  red in IHSUB; desf.
  red; splits; ins; desf; try omega.
  eauto using lt_n_S.
  eauto using lt_n_S.
Qed.

Lemma subseq_mapping_conv (l1 l2 : list A) 
  (f : nat -> nat)
  (MON: forall n1 n2 (LEN: n2 < length l1), n1 < n2 -> f n1 < f n2)
  (NTH: forall n, n < length l1 -> nth_error l1 n = nth_error l2 (f n)):
  subseq l1 l2.
Proof.
generalize dependent f.
generalize dependent l1.
induction l2; ins.
- destruct l1; [by econs|].
specialize (NTH 0 (Nat.lt_0_succ _)).
destruct (f 0); ins.
- ins; destruct (f 0) eqn:F.
+ ins; destruct l1.
by econs.
assert (NTH':=NTH).
specialize (NTH 0 (Nat.lt_0_succ _)); ins; rewrite F in NTH; ins; desf.
apply sub_fst.
eapply IHl2 with (f:= fun n => Nat.pred (f (S n))); ins.
* apply (Nat.pred_lt_mono (f (S n1)) (f (S n2))).
-- intro H0; specialize (MON 0 (S n1)).
   rewrite H0 in MON; eapply Nat.nlt_0_r; eapply MON; omega.
-- apply MON; omega.
* specialize (NTH' (S n)); ins.
rewrite NTH'; try omega.
destruct (f (S n)) eqn:FS; ins; try omega.
specialize (MON 0 (S n)). 
   rewrite FS in MON.
exfalso; eapply Nat.nlt_0_r; eapply MON; omega.
+ ins; destruct l1.
by econs.
apply sub_nfst.
eapply IHl2 with (f:= fun n => Nat.pred (f n)); ins.
* apply (Nat.pred_lt_mono (f n1) (f n2)); eauto.
intro; specialize (MON 0 n1).
destruct (f n1) eqn:FS; ins.
exfalso.
destruct n1; desf.
congruence.
eapply Nat.nlt_0_r; eapply MON; omega.
* specialize (NTH n0); ins.
rewrite NTH; try omega.
destruct (f n0) eqn:FS; ins.
exfalso.
destruct n0; desf.
congruence.
specialize (MON 0 (S n0)).
rewrite FS in MON.
eapply Nat.nlt_0_r; eapply MON; omega.
Qed.

Lemma subseq_witness_nth 
  (l1 l2 : list A) (f : nat -> nat) 
  (SUB: subseq_witness l1 l2 f) : 
  forall n d1 d2, n < length l1 -> nth n l1 d1 = nth (f n) l2 d2.
Proof.
red in SUB; desc.
ins.
apply nth_error_nth.
rewrite NTH; try done.
apply nth_error_nth'; eauto.
Qed.

End Subseq.

(******************************************************************************)
(** * Specific lemma for list of bool  *)
(******************************************************************************)

Lemma subseq_witness_min_of_P
  (l1 l2 : list bool) (f : nat -> nat) 
  (SUB: subseq_witness l1 l2 f) 
  (IN:  min_of_P l1 < length l1) : 
  min_of_P l2 <= f (min_of_P l1).
Proof.
remember (min_of_P l1) as m1.
symmetry in Heqm1.
rewrite <- min_of_P_alt in Heqm1; desf; try edone.
remember (min_of_P l2) as m2.
symmetry in Heqm2.
rewrite <- min_of_P_alt in Heqm2; desf.
- destruct (classic (m2 <= f m1)); auto.
  exfalso.
  apply Nat.nle_gt in H.
  specialize (Heqm3 _ H).
  erewrite <- subseq_witness_nth with (d1:=false) in Heqm3; try edone.
  desf.
- eapply P_non_empty_min.
  exists (f m1).
  rewrite nth_error_nth' with (d:= false).
  erewrite <- subseq_witness_nth with (d1:=false); try edone.
  by f_equal.
  red in SUB; desf; eauto.
Qed.

